@@ -4,7 +4,7 @@ from ihm import AsymUnitRange, _remove_identical  # noqa: F401
 import modelcif.data
 import sys
 
-__version__ = '0.3'
+__version__ = '0.4'
 
 
 class System(object):
@@ -15,6 +15,8 @@ class System(object):
        :param database: If this system is part of an official database
               (e.g. SwissModel, ModBase), details of the database identifiers.
        :type database: :class:`Database`
+       :param str model_details: Detailed description of the system, like an
+                                 abstract.
 
        The system contains a number of simple flat lists of various objects,
        for example :attr:`alignments`. After constructing objects they should
@@ -33,9 +35,11 @@ class System(object):
 
     structure_determination_methodology = "computational"
 
-    def __init__(self, title=None, id='model', database=None):
+    def __init__(self, title=None, id='model', database=None,
+                 model_details=None):
         self.id, self.title = id, title
         self.database = database
+        self.model_details = model_details
 
         #: List of plain text comments. These will be added to the top of
         #: the mmCIF file.
@@ -65,6 +69,10 @@ class System(object):
         #: See :mod:`modelcif.alignment`.
         self.alignments = []
 
+        #: Any additional files with extra data about this system.
+        #: See :class:`modelcif.associated.Repository`.
+        self.repositories = []
+
         self.entities = []
         self.asym_units = []
         self.target_entities = []
@@ -123,7 +131,9 @@ class System(object):
     def _all_templates(self):
         return itertools.chain(
             self.templates,
-            (x.template for x in self.template_segments))
+            (x.template for x in self.template_segments),
+            (x.template for x in self.asym_units
+             if isinstance(x, NonPolymerFromTemplate)))
 
     def _all_template_transformations(self):
         return itertools.chain(
@@ -186,10 +196,10 @@ class System(object):
             self.asym_units, _all_asym_in_assemblies())
 
     def _all_entities(self):
+        # Note that template entities are not included by default
         return itertools.chain(
             self.entities,
-            (asym.entity for asym in self.asym_units),
-            (template.entity for template in self.templates))
+            (asym.entity for asym in self.asym_units))
 
     def _all_model_groups(self, only_in_states=True):
         return self.model_groups
@@ -228,6 +238,10 @@ This can be used both for template sequences (in which case the Entity is
 then used in a :class:`Template` object) or for target (model) sequences
 (where it is used in a :class:`AsymUnit` object).
 
+(Note that template sequence Entity objects are not written out to the
+entity, entity_poly etc. tables in the mmCIF/BinaryCIF file by default.
+Instead, sequence information is captured in template-specific categories.)
+
 :param sequence sequence: The primary sequence, as a sequence of
        :class:`ihm.ChemComp` objects, and/or codes looked up in `alphabet`.
        See `ihm.Entity <https://python-ihm.readthedocs.io/en/latest/main.html#ihm.Entity>`_ for examples.
@@ -383,15 +397,17 @@ class TemplateSegment(object):
         self.gapped_sequence = gapped_sequence
         self.seq_id_range = (seq_id_begin, seq_id_end)
 
-    entity = property(lambda self: self.template.entity)
-
 
 class Template(modelcif.data.Data):
     """A single chain that was used as a template structure for modeling.
 
-       After creating a template, use :meth:`segment` to denote the part of
-       its sequence used in any modeling alignments
+       After creating a polymer template, use :meth:`segment` to denote the
+       part of its sequence used in any modeling alignments
        (see :class:`modelcif.alignment.Pair`).
+
+       Non-polymer templates do not have alignments, and should instead be
+       passed to one or more :class:`NonPolymerFromTemplate` objects.
+
        Template objects can also be used as inputs or outputs in modeling
        protocol steps; see :class:`modelcif.protocol.Step`.
 
@@ -409,11 +425,13 @@ class Template(modelcif.data.Data):
              objects
        :param str strand_id: PDB or "author-provided" strand/chain ID.
               If not specified, it will be the same as the regular asym_id.
+       :param str entity_id: If known, the ID of the entity for this template
+              in its own mmCIF file.
     """
     data_content_type = "template structure"
 
     def __init__(self, entity, asym_id, model_num, transformation,
-                 name=None, references=[], strand_id=None):
+                 name=None, references=[], strand_id=None, entity_id=None):
         super(Template, self).__init__(name)
         self.entity = entity
         self.asym_id, self.model_num = asym_id, model_num
@@ -421,6 +439,7 @@ class Template(modelcif.data.Data):
         self.references = []
         self.references.extend(references)
         self._strand_id = strand_id
+        self.entity_id = entity_id
 
     def segment(self, gapped_sequence, seq_id_begin, seq_id_end):
         """Get an object representing the alignment of part of this sequence.
@@ -439,3 +458,32 @@ class Template(modelcif.data.Data):
 
     strand_id = property(lambda self: self._strand_id or self.asym_id,
                          doc="PDB or author-provided strand/chain ID")
+
+
+class NonPolymerFromTemplate(AsymUnit):
+    """A non-polymer (e.g. ligand) in the model that is modeled from
+       a non-polymer template.
+
+       These objects act just like :class:`AsymUnit` and should be added
+       to :class:`Assembly`.
+
+       To represent a non-polymer that is modeled without a template, just
+       use a regular :class:`AsymUnit`.
+
+       :param template: The non-polymer template used to model
+              this non-polymer.
+       :type template: :class:`Template`
+       :param bool explicit: True iff the conformation of the template is
+              allowed to change (e.g. bond relaxation, flexible fitting)
+              during the modeling, or False if the template is treated as
+              a single rigid body.
+
+       For the other parameters, see :class:`AsymUnit`.
+    """
+
+    def __init__(self, template, explicit, details=None, auth_seq_id_map=0,
+                 id=None, strand_id=None):
+        super(NonPolymerFromTemplate, self).__init__(
+            template.entity, details=details, auth_seq_id_map=auth_seq_id_map,
+            id=id, strand_id=strand_id)
+        self.template, self.explicit = template, explicit
