@@ -6,14 +6,18 @@ import modelcif.protocol
 import modelcif.qa_metric
 import modelcif.alignment
 import modelcif.reference
+import modelcif.associated
 import ihm
 import ihm.source
 import ihm.reader
 from ihm.reader import Variant, Handler, IDMapper, _ChemCompIDMapper
 from ihm.reader import OldFileError, _make_new_entity
+from datetime import date
+import posixpath
 import operator
 import inspect
 import collections
+import functools
 
 
 class _AuditConformHandler(Handler):
@@ -94,6 +98,8 @@ class _SystemReader(object):
         self.protocols = IDMapper(self.system.protocols,
                                   modelcif.protocol.Protocol)
 
+        self.assoc_by_id = {}
+
         self.qa_by_id = {}
 
         self.software_parameters = collections.defaultdict(list)
@@ -108,6 +114,9 @@ class _SystemReader(object):
         # Correspondence between target and template chains
         self.target_asym_for_template = {}
 
+        # Mapping from Entity to bool ma_model_mode flag
+        self.ma_model_mode_map = {}
+
     def finalize(self):
         # make sequence immutable (see also _make_new_entity)
         for e in self.system.entities:
@@ -122,6 +131,63 @@ class _DatabaseHandler(Handler):
             id=database_id, code=database_code)
 
 
+class _EntityNonPolyHandler(Handler):
+    category = '_pdbx_entity_nonpoly'
+    _mmmap = {'explicit': True, 'implicit': False}
+
+    def __call__(self, entity_id, comp_id, ma_model_mode):
+        s = self.sysr.entities.get_by_id(entity_id)
+        s.sequence = (self.sysr.chem_comps.get_by_id(comp_id),)
+        self.sysr.ma_model_mode_map[s] = self._mmmap.get(ma_model_mode.lower())
+
+
+class _TemplatePolyHandler(Handler):
+    category = '_ma_template_poly'
+
+    def __init__(self, sysr):
+        super(_TemplatePolyHandler, self).__init__(sysr)
+        # Use python-ihm's _EntityPolyHandler to do most of the work here
+        self._eph = ihm.reader._EntityPolyHandler(sysr)
+
+    def __call__(self, template_id, seq_one_letter_code,
+                 seq_one_letter_code_can):
+        self._eph(entity_id=template_id, type=None,
+                  pdbx_seq_one_letter_code=seq_one_letter_code,
+                  pdbx_seq_one_letter_code_can=seq_one_letter_code_can)
+
+    def finalize(self):
+        # No type given in _ma_template_poly, so assume amino acid
+        alphabet = ihm.LPeptideAlphabet()
+
+        def _get_seq(codes, codes_can):
+            for i, code in enumerate(codes):
+                # Hopefully non-standard codes are listed in chem_comp table
+                if code in alphabet:
+                    yield alphabet[code]
+                else:
+                    comp = self.sysr.chem_comps.get_by_id(code)
+                    # chem_comp table doesn't define one-letter codes, so
+                    # fill them in here if we have them
+                    comp.code = code
+                    if i < len(codes_can):
+                        comp.code_canonical = codes_can[i]
+                    yield comp
+        for t in self.system.templates:
+            ei = self._eph._entity_info.get(t._id)
+            if ei:
+                t.entity = ihm.Entity(sequence=_get_seq(ei.one_letter,
+                                                        ei.one_letter_can))
+
+
+class _TemplateNonPolyHandler(Handler):
+    category = '_ma_template_non_poly'
+
+    def __call__(self, template_id, comp_id):
+        template = self.sysr.templates.get_by_id(template_id)
+        seq = (self.sysr.chem_comps.get_by_id(comp_id),)
+        template.entity = ihm.Entity(sequence=seq)
+
+
 class _SoftwareGroupHandler(Handler):
     category = '_ma_software_group'
 
@@ -236,16 +302,28 @@ class _TargetRefDBHandler(Handler):
         self.type_map = _EnumerationMapper(modelcif.reference,
                                            modelcif.reference.TargetReference)
 
+    def get_date(self, iso_date_str):
+        """Get a datetime.date obj for a string in isoformat."""
+        if iso_date_str is None:
+            return None
+        return date(int(iso_date_str[0:4]),
+                    int(iso_date_str[5:7]),
+                    int(iso_date_str[8:10]))
+
     def __call__(self, target_entity_id, db_name, db_name_other_details,
                  db_code, db_accession, seq_db_isoform, seq_db_align_begin,
-                 seq_db_align_end, ncbi_taxonomy_id, organism_scientific):
+                 seq_db_align_end, ncbi_taxonomy_id, organism_scientific,
+                 seq_db_sequence_version_date, seq_db_sequence_checksum):
         e = self.sysr.entities.get_by_id(target_entity_id)
         typ = self.type_map.get(db_name, db_name_other_details)
         ref = typ(code=db_code, accession=db_accession,
                   align_begin=self.get_int(seq_db_align_begin),
                   align_end=self.get_int(seq_db_align_end),
                   isoform=seq_db_isoform, ncbi_taxonomy_id=ncbi_taxonomy_id,
-                  organism_scientific=organism_scientific)
+                  organism_scientific=organism_scientific,
+                  sequence_version_date=self.get_date(
+                      seq_db_sequence_version_date),
+                  sequence_crc64=seq_db_sequence_checksum)
         e.references.append(ref)
 
 
@@ -270,8 +348,10 @@ class _TemplateDetailsHandler(Handler):
         template = self.sysr.templates.get_by_id(template_id)
         template.transformation = self.sysr.transformations.get_by_id(
             template_trans_matrix_id)
-        template.entity = self.sysr.entities.get_by_id(
-            template_label_entity_id)
+        # Add empty sequence (hopefully will fill in from _ma_template_poly
+        # or _ma_template_non_poly)
+        template.entity = ihm.Entity([])
+        template.entity_id = template_label_entity_id
         template.asym_id = template_label_asym_id
         template.model_num = self.get_int(template_model_num)
         template._strand_id = template_auth_asym_id
@@ -355,11 +435,9 @@ class _AlignmentInfoHandler(Handler):
         self.sysr.system.alignments.append(alignment)
 
     def finalize(self):
-        seen_pairs = set()
         for aln in self.sysr.system.alignments:
             for pair in self.sysr.alignment_pairs[aln._id]:
                 k = (pair.template._id, pair.target.asym._id)
-                seen_pairs.add(k)
                 pair.target.seq_id_range = \
                     self.sysr.target_template_poly_mapping.get(k)
                 aln.pairs.append(pair)
@@ -369,23 +447,15 @@ class _AlignmentInfoHandler(Handler):
                     aln.pairs[0].template.gapped_sequence = sequence
                 else:  # target
                     aln.pairs[0].target.gapped_sequence = sequence
-        # Handle templates without explicit sequence alignments
-        for tmpl_target in self.sysr.target_asym_for_template.items():
-            if tmpl_target in seen_pairs:
-                continue
-            template_id, target_asym_id = tmpl_target
-            template = self.sysr.templates.get_by_id(template_id)
-            asym = self.sysr.asym_units.get_by_id(target_asym_id)
-            p = modelcif.alignment.Pair(template=template, target=asym,
-                                        identity=None, score=None)
-
-            class Alignment(modelcif.alignment.Global,
-                            modelcif.alignment.Pairwise):
-                pass
-
-            # We have to make a new Alignment since there is no alignment_id
-            aln = Alignment(name="Modeling alignment", pairs=[p])
-            self.sysr.system.alignments.append(aln)
+        # Handle nonpolymer templates
+        for tmpl_id, tgt_asym_id in self.sysr.target_asym_for_template.items():
+            template = self.sysr.templates.get_by_id(tmpl_id)
+            if not template.entity.is_polymeric():
+                asym = self.sysr.asym_units.get_by_id(tgt_asym_id)
+                asym.__class__ = modelcif.NonPolymerFromTemplate
+                asym.template = template
+                asym.explicit = self.sysr.ma_model_mode_map.get(
+                    template.entity)
 
 
 class _AlignmentHandler(Handler):
@@ -452,7 +522,10 @@ class _AssemblyHandler(Handler):
     def __call__(self, assembly_id, asym_id, seq_id_begin, seq_id_end):
         a = self.sysr.assemblies.get_by_id(assembly_id)
         asym = self.sysr.asym_units.get_by_id(asym_id)
-        a.append(asym(int(seq_id_begin), int(seq_id_end)))
+        if seq_id_begin is None and seq_id_end is None:
+            a.append(asym)
+        else:
+            a.append(asym(int(seq_id_begin), int(seq_id_end)))
 
     def finalize(self):
         # Any AsymUnitRange which covers an entire asym,
@@ -542,6 +615,89 @@ class _ProtocolHandler(Handler):
         p.steps.append(step)
 
 
+def _get_assoc_type_maps():
+    # Get a mapping from (file_content,file_format) to a subclass of
+    # modelcif.associated.File
+    cs = [x[1] for x in inspect.getmembers(modelcif.associated,
+                                           inspect.isclass)
+          if issubclass(x[1], modelcif.associated.File)
+          and x[1] is not modelcif.associated.File]
+    _type_map = dict(
+        ((x.file_content.upper(), x.file_format.upper()), x)
+        for x in cs if not hasattr(x, '_binary_ff_map'))
+    # Do the same thing for classes that take a 'binary' argument
+    _bin_type_map = {}
+    for x in cs:
+        if not hasattr(x, '_binary_ff_map'):
+            continue
+        file_content = x.file_content.upper()
+        for binary, file_format in x._binary_ff_map.items():
+            _bin_type_map[(file_content, file_format.upper())] = (x, binary)
+    return _type_map, _bin_type_map
+
+
+def _get_assoc_class(file_content, file_format, type_map, binary_type_map):
+    # Use previous mapping to get a subclass of modelcif.associated.File
+    # from (file_content, file_format)
+    k = (file_content.upper(), file_format.upper())
+
+    filecls_bin = binary_type_map.get(k)
+    if filecls_bin:
+        filecls, binary = filecls_bin
+        return functools.partial(filecls, binary=binary)
+    else:
+        return type_map.get(k, modelcif.associated.File)
+
+
+class _AssociatedHandler(Handler):
+    category = '_ma_entry_associated_files'
+
+    def __init__(self, *args):
+        super(_AssociatedHandler, self).__init__(*args)
+        self._repos_by_root = {}
+        self._type_map, self._binary_type_map = _get_assoc_type_maps()
+
+    def __call__(self, id, file_url, file_type, file_format, file_content,
+                 details):
+        filecls = _get_assoc_class(
+            file_content, file_format, self._type_map, self._binary_type_map)
+        # Assume everything before last slash (if any) is URL root
+        url_root, path = posixpath.split(file_url)
+        url_root = url_root or None
+        r = self._repos_by_root.get(url_root)
+        if not r:
+            r = modelcif.associated.Repository(url_root=url_root, files=[])
+            self._repos_by_root[url_root] = r
+            self.system.repositories.append(r)
+        c = filecls(path=path, details=details)
+        r.files.append(c)
+        self.sysr.assoc_by_id[id] = c
+
+
+class _AssociatedArchiveHandler(Handler):
+    category = '_ma_associated_archive_file_details'
+
+    def __init__(self, *args):
+        super(_AssociatedArchiveHandler, self).__init__(*args)
+        self._type_map, self._binary_type_map = _get_assoc_type_maps()
+        self._archive_files = collections.defaultdict(list)
+
+    def __call__(self, id, archive_file_id, file_path, file_format,
+                 file_content, description):
+        filecls = _get_assoc_class(
+            file_content, file_format, self._type_map, self._binary_type_map)
+        c = filecls(path=file_path, details=description)
+        # Top-level archive file might not exist yet
+        self._archive_files[archive_file_id].append(c)
+
+    def finalize(self):
+        # Put files in archives
+        for archive_file_id, files in self._archive_files.items():
+            archive = self.sysr.assoc_by_id.get(archive_file_id)
+            if archive:
+                archive.files = files
+
+
 def _make_qa_class(type_class, mode_class, p_name, p_description, p_software):
     """Create and return a new class to represent a QA metric"""
     class QA(type_class, mode_class):
@@ -631,16 +787,18 @@ class ModelCIFVariant(Variant):
         ihm.reader._ChemCompHandler, ihm.reader._EntityHandler,
         ihm.reader._EntitySrcNatHandler, ihm.reader._EntitySrcGenHandler,
         ihm.reader._EntitySrcSynHandler, ihm.reader._EntityPolyHandler,
-        ihm.reader._EntityPolySeqHandler, ihm.reader._EntityNonPolyHandler,
+        ihm.reader._EntityPolySeqHandler, _EntityNonPolyHandler,
         ihm.reader._StructAsymHandler, _SoftwareGroupHandler,
         _DatabaseHandler, _SoftwareParameterHandler,
         _DataHandler, _DataGroupHandler, _TargetEntityHandler,
         _TargetRefDBHandler, _TransformationHandler, _TemplateDetailsHandler,
         _TemplateRefDBHandler, _TemplatePolySegmentHandler,
+        _TemplatePolyHandler, _TemplateNonPolyHandler,
         _AlignmentHandler, _AlignmentInfoHandler, _AlignmentDetailsHandler,
         _TargetTemplatePolyMappingHandler,
         _AssemblyHandler, _AssemblyDetailsHandler, ihm.reader._AtomSiteHandler,
-        _ModelListHandler, _ProtocolHandler, _QAMetricHandler,
+        _ModelListHandler, _ProtocolHandler,
+        _AssociatedHandler, _AssociatedArchiveHandler, _QAMetricHandler,
         _QAMetricGlobalHandler, _QAMetricLocalHandler,
         _QAMetricPairwiseHandler]
 
@@ -665,6 +823,10 @@ def read(fh, model_class=modelcif.model.Model, format='mmCIF', handlers=[],
        stated in the mmCIF file (e.g. homology model, ab initio model).
        (However, the ``model_type`` attribute will be set appropriately.)
 
+       If the input file references any associated files, they will be
+       listed in :attr:`modelcif.System.repositories`. The files will not be
+       downloaded or read in, however.
+
       :return: A list of :class:`modelcif.System` objects.
     """
     return ihm.reader.read(
