@@ -1,3 +1,4 @@
+from datetime import date
 import sys
 import unittest
 import utils
@@ -152,10 +153,13 @@ _ma_target_ref_db_details.seq_db_align_begin
 _ma_target_ref_db_details.seq_db_align_end
 _ma_target_ref_db_details.ncbi_taxonomy_id
 _ma_target_ref_db_details.organism_scientific
-1 UNP . MED1_YEAST Q12321 test_iso 1 10 test_tax test_org
-1 Other foo . . ? 1 10 . .
-1 other bar . . ? 1 10 . .
-1 MIS baz . . ? 1 10 . .
+_ma_target_ref_db_details.seq_db_sequence_version_date
+_ma_target_ref_db_details.seq_db_sequence_checksum
+1 UNP . MED1_YEAST Q12321 test_iso 1 10 test_tax test_org 1996-11-01
+637FEA3E78D915BC
+1 Other foo . . ? 1 10 . . . .
+1 other bar . . ? 1 10 . . . .
+1 MIS baz . . ? 1 10 . . . .
 """
         s, = modelcif.reader.read(StringIO(cif))
         e, = s.entities
@@ -168,6 +172,7 @@ _ma_target_ref_db_details.organism_scientific
         self.assertEqual(r1.align_end, 10)
         self.assertEqual(r1.ncbi_taxonomy_id, 'test_tax')
         self.assertEqual(r1.organism_scientific, 'test_org')
+        self.assertEqual(r1.sequence_version_date, date(1996, 11, 1))
         self.assertEqual(r2.name, 'Other')
         self.assertEqual(r2.other_details, 'foo')
         self.assertEqual(r3.name, 'Other')
@@ -216,20 +221,57 @@ _ma_template_details.template_label_entity_id
 _ma_template_details.template_model_num
 _ma_template_details.template_auth_asym_id
 1 1 'reference database' polymer 1 2 A B 3 4 Z
+2 2 'reference database' polymer 2 3 . B 3 4 Z
 """
         s, = modelcif.reader.read(StringIO(cif))
-        t, = s.templates
-        self.assertEqual(t.entity._id, '3')
+        t, t2 = s.templates
+        self.assertEqual(t.entity_id, '3')
         self.assertEqual(t.model_num, 4)
         self.assertEqual(t.asym_id, 'B')
         self.assertEqual(t.strand_id, 'Z')
-        # An 'alignment' should have been created for the implied
-        # template - target_asym_id correspondence
-        a, = s.alignments
-        p, = a.pairs
-        self.assertEqual(p.template._id, '1')
-        self.assertIs(p.template, t)
-        self.assertEqual(p.target._id, 'A')
+        self.assertEqual(len(s.alignments), 0)
+
+    def test_template_details_handler_nonpoly(self):
+        """Test _TemplateDetailsHandler with nonpolymeric template"""
+        cif = """
+loop_
+_pdbx_entity_nonpoly.entity_id
+_pdbx_entity_nonpoly.name
+_pdbx_entity_nonpoly.comp_id
+_pdbx_entity_nonpoly.ma_model_mode
+3 Heme HEM explicit
+#
+loop_
+_ma_template_details.ordinal_id
+_ma_template_details.template_id
+_ma_template_details.template_origin
+_ma_template_details.template_entity_type
+_ma_template_details.template_trans_matrix_id
+_ma_template_details.template_data_id
+_ma_template_details.target_asym_id
+_ma_template_details.template_label_asym_id
+_ma_template_details.template_label_entity_id
+_ma_template_details.template_model_num
+_ma_template_details.template_auth_asym_id
+1 1 'reference database' non-polymer 1 2 A B 3 4 Z
+#
+loop_
+_ma_template_non_poly.template_id
+_ma_template_non_poly.comp_id
+_ma_template_non_poly.details
+1 HEM Heme
+"""
+        s, = modelcif.reader.read(StringIO(cif))
+        t, = s.templates
+        self.assertEqual(t.entity_id, '3')
+        self.assertEqual(t.model_num, 4)
+        self.assertEqual(t.asym_id, 'B')
+        self.assertEqual(t.strand_id, 'Z')
+        self.assertEqual(len(s.alignments), 0)
+        a, = s.asym_units
+        self.assertIsInstance(a, modelcif.NonPolymerFromTemplate)
+        self.assertIs(a.template, t)
+        self.assertTrue(a.explicit)
 
     def test_template_ref_db_handler(self):
         """Test _TemplateRefDBHandler"""
@@ -774,6 +816,157 @@ _ma_target_template_poly_mapping.target_seq_id_end
         self.assertEqual(p.target.gapped_sequence, 'DSYV-ETLD')
         self.assertEqual(p.target.seq_id_range, (1, 8))
 
+    def test_associated_files(self):
+        """Test _AssociatedHandler and _AssociatedArchiveHandler"""
+        cif = """
+loop_
+_ma_entry_associated_files.id
+_ma_entry_associated_files.entry_id
+_ma_entry_associated_files.file_url
+_ma_entry_associated_files.file_type
+_ma_entry_associated_files.file_format
+_ma_entry_associated_files.file_content
+_ma_entry_associated_files.details
+1 model https://example.com/foo.txt file other other 'test file'
+2 model https://example.com/t.zip archive zip 'archive with multiple files' .
+3 model baz.txt file other other 'test file3'
+4 model baz.cif file cif other 'test mmCIF'
+5 model baz.bcif file bcif other 'test BinaryCIF'
+#
+#
+loop_
+_ma_associated_archive_file_details.id
+_ma_associated_archive_file_details.archive_file_id
+_ma_associated_archive_file_details.file_path
+_ma_associated_archive_file_details.file_format
+_ma_associated_archive_file_details.file_content
+_ma_associated_archive_file_details.description
+1 2 bar.txt other other 'test file2'
+2 99 99.txt other other 'test file99'
+3 2 bar.cif cif other 'test mmCIF in zip'
+4 2 bar.bcif bcif 'local pairwise QA scores' 'test BinaryCIF in zip'
+"""
+        s, = modelcif.reader.read(StringIO(cif))
+        r1, r2 = s.repositories
+        self.assertEqual(r1.url_root, 'https://example.com')
+        f1, zf = r1.files
+        self.assertIsInstance(f1, modelcif.associated.File)
+        self.assertEqual(f1.path, 'foo.txt')
+        self.assertEqual(f1.details, 'test file')
+
+        self.assertIsInstance(zf, modelcif.associated.ZipFile)
+        self.assertEqual(zf.path, 't.zip')
+        self.assertIsNone(zf.details)
+
+        f2, f3, f4 = zf.files
+        self.assertEqual(f2.path, 'bar.txt')
+        self.assertEqual(f2.details, 'test file2')
+        self.assertIsInstance(f3, modelcif.associated.CIFFile)
+        self.assertFalse(f3.binary)
+        self.assertIsInstance(
+            f4, modelcif.associated.LocalPairwiseQAScoresFile)
+        self.assertTrue(f4.binary)
+
+        self.assertIsNone(r2.url_root)
+        f3, f4, f5 = r2.files
+        self.assertEqual(f3.path, 'baz.txt')
+        self.assertEqual(f3.details, 'test file3')
+        self.assertIsInstance(f4, modelcif.associated.CIFFile)
+        self.assertFalse(f4.binary)
+        self.assertIsInstance(f5, modelcif.associated.CIFFile)
+        self.assertTrue(f5.binary)
+
+    def test_template_poly_handler(self):
+        """Test _TemplatePolyHandler"""
+        cif = """
+loop_
+_chem_comp.id
+_chem_comp.type
+_chem_comp.name
+_chem_comp.formula
+MYTYPE 'D-PEPTIDE LINKING' 'MY CUSTOM COMPONENT' 'C6 H12'
+MYTYP2 'D-PEPTIDE LINKING' 'MY CUSTOM COMPONENT2' 'C6 H12'
+#
+loop_
+_ma_template_details.ordinal_id
+_ma_template_details.template_id
+_ma_template_details.template_origin
+_ma_template_details.template_entity_type
+_ma_template_details.template_trans_matrix_id
+_ma_template_details.template_data_id
+_ma_template_details.target_asym_id
+_ma_template_details.template_label_asym_id
+_ma_template_details.template_label_entity_id
+_ma_template_details.template_model_num
+_ma_template_details.template_auth_asym_id
+1 1 'reference database' polymer 1 1 A A . 1 A
+2 2 'reference database' polymer 1 1 A A . 1 A
+3 3 'reference database' polymer 1 1 A A . 1 A
+#
+loop_
+_ma_template_poly.template_id
+_ma_template_poly.seq_one_letter_code
+_ma_template_poly.seq_one_letter_code_can
+1 A(MYTYPE)V AVV
+2 A(MYTYP2)V .
+3 . .
+4 CCC CCC
+"""
+        s, = modelcif.reader.read(StringIO(cif))
+        # template_id=4 in template_poly should be ignored
+        t1, t2, t3 = s.templates
+        s1, s2, s3 = t1.entity.sequence
+        self.assertEqual(s1.id, 'ALA')
+        self.assertEqual(s1.code, 'A')
+        # Both one-letter and one-letter-canonical were provided
+        self.assertEqual(s2.id, 'MYTYPE')
+        self.assertEqual(s2.code, 'MYTYPE')
+        self.assertEqual(s2.code_canonical, 'V')
+
+        # Only one-letter was provided
+        s1, s2, s3 = t2.entity.sequence
+        self.assertEqual(s2.id, 'MYTYP2')
+        self.assertEqual(s2.code, 'MYTYP2')
+        self.assertIsNone(s2.code_canonical)
+
+        # No sequence provided
+        self.assertEqual(len(t3.entity.sequence), 0)
+
+    def test_template_non_poly_handler(self):
+        """Test _TemplateNonPolyHandler"""
+        cif = """
+loop_
+_chem_comp.id
+_chem_comp.type
+HEM non-polymer
+#
+loop_
+_ma_template_details.ordinal_id
+_ma_template_details.template_id
+_ma_template_details.template_origin
+_ma_template_details.template_entity_type
+_ma_template_details.template_trans_matrix_id
+_ma_template_details.template_data_id
+_ma_template_details.target_asym_id
+_ma_template_details.template_label_asym_id
+_ma_template_details.template_label_entity_id
+_ma_template_details.template_model_num
+_ma_template_details.template_auth_asym_id
+1 1 'reference database' polymer 1 1 A A . 1 A
+#
+loop_
+_ma_template_non_poly.template_id
+_ma_template_non_poly.comp_id
+_ma_template_non_poly.details
+1 HEM Heme
+"""
+        s, = modelcif.reader.read(StringIO(cif))
+        t, = s.templates
+        s1, = t.entity.sequence
+        self.assertEqual(s1.id, 'HEM')
+        self.assertEqual(s1.type, 'non-polymer')
+        self.assertIsInstance(s1, ihm.NonPolymerChemComp)
+
 
 if __name__ == '__main__':
     unittest.main()
